(* $Id: ugraphics.ml,v 1.12 2014/09/08 12:41:48 fclement Exp $ *)

let _0_ = Unatural._0_
and _1_ = Unatural._1_
and _2_ = Unatural._2_
;;

type coordinate = Unatural.t;;

type abscissa = coordinate
and ordinate = coordinate
;;

type center_abscissa = abscissa
and center_ordinate = ordinate
;;

type length = Unatural.t;;

type width = length
and height = length
;;

type aspect_ratio = width * height;;

type radius = length;;

type principal_radius = radius
and secondary_radius = radius
;;

type angle = int;;

type from_angle = angle
and to_angle = angle
;;

type color = int32;;

type color_channel = char;;

type alpha_channel = color_channel
and red_channel = color_channel
and green_channel = color_channel
and blue_channel = color_channel
;;

type from_color = color
and to_color = color
;;

let aspect_ratio_of_int_pair (w, h) =
  let w = Unatural.of_int w in
  let h = Unatural.of_int h in
  (w, h)
;;

let int_pair_of_aspect_ratio (w, h) =
  let w = Unatural.to_int w in
  let h = Unatural.to_int h in
  (w, h)
;;

let int32_of_char c = Int32.of_int (int_of_char c);;

let char_of_int32 i = char_of_int (Int32.to_int (Int32.logand i 0x000000ffl));;

let color_channel_of_natural = Unatural.to_char
let natural_of_color_channel = Unatural.of_char;;

let a_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0xff000000l) 24)
and r_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0x00ff0000l) 16)
and g_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0x0000ff00l) 8)
and b_of c =
  char_of_int32 (Int32.logand c 0x000000ffl)
;;

let rgb_of_color c = (r_of c, g_of c, b_of c);;

let argb a r g b =
  let ( + ) = Int32.add in
  let ( lsl ) = Int32.shift_left in
  int32_of_char a lsl 24 +
  int32_of_char r lsl 16 +
  int32_of_char g lsl 8 +
  int32_of_char b
;;

let rgb = argb '\x00';;

let rgb_int r g b = rgb (char_of_int r) (char_of_int g) (char_of_int b);;

let red = rgb '\xFF' '\x00' '\x00'
and green = rgb '\x00' '\xFF' '\x00'
and blue = rgb '\x00' '\x00' '\xFF'
and yellow = rgb '\xFF' '\xFF' '\x00'
and cyan = rgb '\x00' '\xFF' '\xFF'
and magenta = rgb '\xFF' '\x00' '\xFF'
and white = rgb '\xFF' '\xFF' '\xFF'
and black = rgb '\x00' '\x00' '\x00'
;;

type steps = Unatural.t
and max_steps = Unatural.t
and from_natural = Unatural.t
and to_natural = Unatural.t
;;

(* Scaling from c to c', by i steps over w steps.

 Meaning:
  scale 0 w c c' = c           (1)
  scale w w c c' = c'          (2)
  scale i w c c' = a * i + b   (3) for some real a and b.

  We have  b = c               (due to (1))
  and      c'= a * w + c       (due to (2))
  hence    a = (c' - c) / w

  Finally, we get:

  scale i w c c' = (c' * i + c * (w - i)) / w
*)
let scale i w c c' =
  if Unatural.is_zero w then c else
  let i = Unatural.to_int i
  and w = Unatural.to_int w
  and c = Unatural.to_int c
  and c' = Unatural.to_int c' in
  let mid = (w + 1) / 2 in
  let round_quot w x = (x + mid) / w in
  Unatural.of_int (round_quot w (c' * i + c * (w - i)))
;;

let round_float f = Unatural.of_float (f +. 0.5);;

let scale_natural f n = round_float (f *. Unatural.to_float n);;

(* Scaling a color from c to c', by i steps over w steps. *)
let scale_color i w c c' =
  assert (i <= w);
  argb
   (a_of c)
   (color_channel_of_natural
      (scale i w
        (Unatural.of_char (r_of c))
        (Unatural.of_char (r_of c'))))
   (color_channel_of_natural
      (scale i w
        (Unatural.of_char (g_of c))
        (Unatural.of_char (g_of c'))))
   (color_channel_of_natural
      (scale i w
        (Unatural.of_char (b_of c))
        (Unatural.of_char (b_of c'))))
;;

let fill_rectangle x y w h =
  if Unatural.is_positive w && Unatural.is_positive h then
    let x = Unatural.to_int x
    and y = Unatural.to_int y
    and w = Unatural.to_int (Unatural.pred w)
    and h = Unatural.to_int (Unatural.pred h) in
    Graphics.fill_rect x y w h
;;

let diameter r = Unatural.add (Unatural.mul _2_ r) _1_;;

let fill_rectangle_ball x y rx ry =
  fill_rectangle
    (Unatural.sub x rx) (Unatural.sub y ry)
    (diameter rx) (diameter ry)
;;

let fill_ellipse_ball x y rx ry =
  if Unatural.is_zero rx || Unatural.is_zero ry then
    fill_rectangle_ball x y rx ry else
  let x = Unatural.to_int x
  and y = Unatural.to_int y
  and rx = Unatural.to_int rx
  and ry = Unatural.to_int ry in
  Graphics.fill_ellipse x y rx ry
;;

let radius d = Unatural.div (Unatural.sub d _1_) _2_;;

let fill_inscripted_ellipse x y w h =
  if Unatural.is_odd w && Unatural.is_odd h then
    let rx = radius w in
    let ry = radius h in
    let xc = Unatural.add x rx in
    let yc = Unatural.add y ry in
    fill_ellipse_ball xc yc rx ry else
  (* Using round is simply wrong! *)
  failwith "fill_inscripted_ellipse: not yet implemented."
;;

(* To fill balls associated with the L_infinite norm, i.e. the sup norm:
  ||u||_sup = max |u_i|. *)
let fill_ball_inf = fill_rectangle_ball;;

(* To fill balls associated with the L_2 norm, i.e. the Euclidean norm:
  ||u||_2 = (sum_i |u_i|^2)^(1/2). *)
let fill_ball_euclidean = fill_ellipse_ball;;

(* Obsolete definitions for backward compatibility.
  Do not use in fresh code. *)
let fill_rect = fill_rectangle;;
let fill_ellipse = fill_ellipse_ball;;

type x = abscissa
and y = ordinate
and xc = center_abscissa
and yc = center_ordinate
and w = width
and h = height
and r = radius
and rx = principal_radius
and ry = secondary_radius
;;

(* Converting to the new types the Graphics basic stuff.
   This should be reimplemented here.
*)
let set_color c =
  let c = Int32.to_int c in
  Graphics.set_color c
;;

let moveto x y =
  let x = Unatural.to_int x
  and y = Unatural.to_int y in
  Graphics.moveto x y
;;

let lineto x y =
  let x = Unatural.to_int x
  and y = Unatural.to_int y in
  Graphics.lineto x y
;;

let set_line_width w =
  let w = Unatural.to_int w in
  Graphics.set_line_width w
;;

let draw_circle x y r =
  let x = Unatural.to_int x
  and y = Unatural.to_int y
  and r = Unatural.to_int r in
  Graphics.draw_circle x y r
;;

let draw_arc x y rx ry a1 a2 =
  let x = Unatural.to_int x
  and y = Unatural.to_int y
  and rx = Unatural.to_int rx
  and ry = Unatural.to_int ry in
  Graphics.draw_arc x y rx ry a1 a2
;;

let plot x y =
  let x = Unatural.to_int x
  and y = Unatural.to_int y in
  Graphics.plot x y
;;

let open_graph = Graphics.open_graph;;

let fill_poly a =
  Graphics.fill_poly
    (Array.map (fun (x, y) -> (Unatural.to_int x, Unatural.to_int y)) a)
;;

let fill_circle x y r =
  let x = Unatural.to_int x
  and y = Unatural.to_int y
  and r = Unatural.to_int r in
  Graphics.fill_circle x y r
;;
