(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2000-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: colors.mli,v 1.19 2017/05/31 14:42:47 fclement Exp $ *)

(** {3 Definitions and manipulations of colors.} *)

type t = Graphics.color;;

(** {6 The RGB color encoding.} *)

(** RGB color encoding, using color decomposition into red, green, and blue
    components. Each component has a value in the range [0..255]. The three
    components are packed together into an integer: [0xRRGGBB], where [RR]
    are the two hexadecimal digits for the red component, [GG] for the green
    component, [BB] for the blue component. *)

type rgb_channel = int;;

type red = rgb_channel
and green = rgb_channel
and blue = rgb_channel
;;

val color_of_rgb : red -> green -> blue -> t
(** [color_of_rgb r g b] returns the integer encoding a color, by packing its
   [r] red component, its [g] green component, and its [b] blue component
   into an integer. *)
;;
val rgb : red -> green -> blue -> t
(** [rgb] is an alias for [color_of_rgb]. *)
;;

val red_of_color : t -> red;;
val green_of_color : t -> green;;
val blue_of_color : t -> blue;;
val rgb_of_color : t -> red * green * blue
(** Extracting the red, green, and blue components of a color. *)
;;

(** {6 The HSB color encoding.} *)

(** HSB color encoding, using color decomposition into hue, saturation, and
  brightness components.
  Hue is the wavelength within the visible-light spectrum at which the
  energy output from a source of this color is the greatest.
  Saturation measures the purity of colors: high saturation means a
  pure color, low saturation means a faint color.
  Brightness (also called brilliance) is the maximum of the RGB
  components of the color.
  As for RGB encoding, each component has a value in the range [0..255] and the
  three components are packed into an integer. *)

type hsb_channel = int;;

type hue = hsb_channel
and saturation = hsb_channel
and brightness = hsb_channel
;;

val hue_of_color : t -> hue;;
val saturation_of_color : t -> saturation;;
val brightness_of_color : t -> brightness;;
val hsb_of_color : t -> hue * saturation * brightness
(** Decomposition of a color into its hue, saturation, and brightness
   components. *)
;;

val color_of_hsb : hue -> saturation -> brightness -> t
(** [color_of_hsb h s b] returns the color with correspondant HSB
  components. *)
;;
val hsb : hue -> saturation -> brightness -> t
(** [hsb] is an alias for [color_of_hsb]. *)
;;

(** {6 Color encoding conversions.} *)

val hsb_of_rgb : red -> green -> blue -> hue * saturation * brightness;;
val rgb_of_hsb : hue -> saturation -> brightness -> red * green * blue
(** Mappings beween RGB and HSB representations of colors. [hsb_of_rgb] and
  [rgb_of_hsb] are not strictly reciprocal, due to limit conditions, where
  some components are undefined and arbitrarily set to 0. *)
;;

(** {6 Shorthands to access color components.} *)

 (** Abbreviations for accessing color components. *)
val r_of : t -> red;;
val g_of : t -> green;;
val b_of : t -> blue;;
val hue_of : t -> hue;;
val sat_of : t -> saturation;;
val bri_of : t -> brightness
(** Return the red, green, blue, hue, saturation, and brightness
  of a color. *)
;;

(** {6 Color transformations} *)

val hue : t -> t;;

type shift = int;;

val shift_hue_of_hsb : shift -> hue -> hue;;
val shift_hue : shift -> t -> t;;

(** Grayscale transformations *)

(** Grayscale transformations.
  Intensity (as in the HSI model) is the average of red, green, and blue
  components.
  Brightness (or value, as in the HSB/HSV model) is the maximum of red,
  green, and blue components.
  Lightness (as in the HSL model) is the average between the minimum and
  the maximum of red, green, and blue components.
  Luma is a weighted average of red, green, and blue components,
  respectively roughly with weights 30%, 59%, and 11% to reflect human
  perception of colors. *)

type gray = int;;

val red_gray : t -> t;;
val green_gray : t -> t;;
val blue_gray : t -> t;;
val hue_gray : t -> t;;
val saturation_gray : t -> t;;
val brightness_gray_of_rgb : red -> green -> blue -> gray;;
val intensity_gray_of_rgb : red -> green -> blue -> gray;;
val lightness_gray_of_rgb : red -> green -> blue -> gray;;
val luma_gray_of_rgb : red -> green -> blue -> gray;;
val brightness_gray : t -> t;;
val intensity_gray : t -> t;;
val lightness_gray : t -> t;;
val luma_gray : t -> t;;

(** Hue transformations *)

(** Hues are 255-periodic integers. *)

type hue_range = hue * hue
(** Hues are 255-periodic integers.
  Hue ranges must be valid integer ranges, ie the right bound must be
  greater or equal to the left bound, and the amplitude must be smaller or
  equal to 255. By convention, the left bound belongs to \[0..255\]. *)
;;

type in_hue_range = hue_range
and out_hue_range = hue_range
;;

val select_color_of_rgb :
  in_hue_range * out_hue_range -> (red -> green -> blue -> gray) ->
  red -> green -> blue -> red * green * blue
;;
val select_color : in_hue_range * out_hue_range -> (t -> t) -> t -> t
 (** Grayscale transformation, except when the hue of the input RGB color
   is in the specified input hue range, it is then linearly mapped to the
   output hue range.
   Raise [Invalid_arg] if a hue range is empty, or has an amplitude larger
   than 255. *)
;;

(**/**)
(** {6 Undocumented functions} *)

val linear_integer_map : in_hue_range* out_hue_range -> int -> int;;
val linear_hue_map : in_hue_range* out_hue_range -> hue -> hue;;
