(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: model.ml,v 1.23 2017/11/07 23:47:13 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module S = Ustring;;
module A = Uarray;;
module M = Umatrix;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module Dbg = Udebug;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;

open Misc_sprint;;

(* Shortcuts for the Refinement library. *)

open Optimization_status_types;;

(* OCaml identity model. *)

let qual = P.qualify "Model";;
let invalid_arg = P.invalid_arg_qual qual;;
let failwith = P.failwith_qual qual;;

(* Abtract references. *)

let init_dim = _0_;;

let { Ar.get = get_dim; set = set_dim; } =
  Ar.make_aref_unique "dimension" init_dim
;;

let init_data_image = M.of_array_array [||];;

let { Ar.get = get_data_image; set = set_data_image; } =
  Ar.make_aref_unique "data image" init_data_image
;;

let init_result_image_file_name = "";;

let { Ar.get = get_result_image_file_name;
      set = set_result_image_file_name; } =
  Ar.make_aref_unique "result image file name"
    init_result_image_file_name
;;

let { Ar.get_flag = get_use_mask; set_flag = set_use_mask; _; } =
  Ar.make_flag ()
;;

(* Image stuff. *)

let sprint_color_matrix = sprint_matrix Colors_print.printer_t;;

let load_data_image fname =
  let dim =
    match Ppm.ppm_magic_number fname with
    | Ppm.P1 | Ppm.P2 | Ppm.P4 | Ppm.P5 -> _1_
    | Ppm.P3 | Ppm.P6 -> _3_ in
  set_dim dim;
  let pixmap = Ppm.load_pixmap fname in
  let idata = M.of_array_array pixmap in
  let (h, w) = M.size idata in
  if h = _0_ || w = _0_ then
    invalid_arg "load_data_image" "empty image" else
  set_data_image idata;
;;

let float_triple_of_color c =
  let (r, g, b) = (Colors.r_of c, Colors.g_of c, Colors.b_of c) in
  (float_of_int r, float_of_int g, float_of_int b)
;;

let of_color dim c =
  let (r, g, b) = float_triple_of_color c in
  if dim = _1_ then
    begin
      if r = g && g = b then [| r; |] else
      failwith "of_color" "color is not grayscale"
    end
  else if dim = _3_ then [| r; g; b; |]
  else assert false
;;

let float_matrix_of_image dim img =
  let im = M.map (of_color dim) img in
  M.transpose (M.of_array_array (M.to_array im))
;;

let float_triple_to_color (r, g, b) =
  Colors.color_of_rgb (int_of_float r) (int_of_float g) (int_of_float b)
;;

let to_color dim a =
  if A.length a <> dim then
    invalid_arg "to_color" "wrong dimension";
  let (r, g, b) =
    if dim = _1_ then
      let a0 = A.unsafe_get a _0_ in
      (a0, a0, a0) else
    if dim = _3_ then
      let a0 = A.unsafe_get a _0_
      and a1 = A.unsafe_get a _1_
      and a2 = A.unsafe_get a _2_ in
      (a0, a1, a2) else
    assert false in
  float_triple_to_color (r, g, b)
;;

let image_of_float_matrix dim (ni, nj) fm =
  let im =
    A_ng.matrix_of_array ni nj
      (M.to_array_array (M.transpose fm)) in
  M.map (to_color dim) im
;;

(* Command-line options. *)

let result_image_file_name_rco =
  {
   Co.key =
     Ok.long_option_key_of_string "result parameter file name";
   spec = Arg.String set_result_image_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet result parameter file name to <fname>.";
  }
;;

let use_mask_rco =
  {
   Co.key = Ok.long_option_key_of_string "use mask";
   spec = Arg.Unit set_use_mask;
   doc = Printf.sprintf "\n\tSet mask usage on. Default is off.";
  }
;;

let add_options () =
  Co.application_add_all_rco
    [ result_image_file_name_rco; use_mask_rco; ]
;;

add_options ();;

let sys_argv_0 = A.get Sys.argv _0_;;

Co.set_usage_msg
  (Format.sprintf
     "\n%s -- Identity worker\n\n\
      Usage: %s [<option>...] <input_image_file_name>\n\n\
      Options are:@."
     sys_argv_0 sys_argv_0)
;;

Co.set_anon_fun load_data_image;;

(* Mask. *)

let selection_matrix (ni, nj) =
  if not (get_use_mask ()) then M.make ni nj P.Keep else
  M.init ni nj
    (fun i j ->
      if Nat.rem (Nat.add i j) _2_ = _0_ then P.Keep else
      P.Forget)
;;

let masking sa m =
  M.of_array_array
    (A.map (A.select sa) (M.to_array_array m))
;;

let unmasking x sa fm =
  M.of_array_array
    (A.map (A.unselect x sa) (M.to_array_array fm))
;;

(* Task engines. *)

let init_data_height = _1_
and init_data_width = _1_
and init_selection_array = A.empty
;;
let init_data = M.make init_data_height init_data_width 0.;;

let { Ar.get = get_data_height; set = set_data_height; } =
  Ar.make_aref_unique "data height" init_data_height
and { Ar.get = get_data_width; set = set_data_width; } =
  Ar.make_aref_unique "data width" init_data_width
and { Ar.get = get_selection_array; set = set_selection_array; } =
  Ar.make_aref_unique "selection array" init_selection_array
and { Ar.get = get_data; set = set_data; } =
  Ar.make_aref_unique "data" init_data
;;

let set_size_and_data (ni, nj) sa dm =
  set_data_height ni;
  set_data_width nj;
  set_selection_array sa;
  set_data dm;
;;

let check_initialized tname f x =
  try f x with
  | Failure s when S.check_prefix "Fatal error in Uaref" s ->
      raise (Pio.Uninitialized_worker tname)
;;

(* Init. *)

let forget_is_minus_one sa =
  fun _i j mij ->
    match A.get sa j with
    | P.Forget -> -1
    | P.Keep -> Nat.to_int mij
;;

let forget_or_keep sa =
  if get_use_mask () then forget_is_minus_one sa else
  (fun _ _ -> Nat.to_int)
;;

(* cvd = check_valid_dimension *)
let init_engine cvd { Dbg.spf_debug = debug; _; } dim =
  if dim <> _1_ && dim <> _3_ then
    invalid_arg "init" "dimension must be 1 or 3";
  let dim' = get_dim () in
  cvd "init" dim dim';
  let img = get_data_image () in
  debug (sprint_color_matrix "img" img);
  let (ni, nj as sz) = M.size img in
  debug (sprint_nat_couple "sz" sz);
  let sm = selection_matrix sz in
  let sa = M.to_array sm in
  debug (sprint_select_array "sa" sa);
  let fm = float_matrix_of_image dim img in
  debug (sprint_float_matrix "fm" fm);
  let dm = masking sa fm in
  debug (sprint_float_matrix "dm" dm);
  set_size_and_data sz sa dm;
  let ppm1 = M.make dim (Nat.mul ni nj) _0_ in
  let mpm1 = M.map_ij (forget_or_keep sa) ppm1 in
  debug (sprint_int_matrix "mpm1" mpm1);
  let ppm1 = masking sa ppm1 in
  debug (sprint_nat_matrix "ppm1" ppm1);
  let maa1_init = A_ng.array_array_make dim (fun _i -> _1_) 0. in
  debug (sprint_float_array_array "maa1_init" maa1_init);
  (sz, mpm1, ppm1, maa1_init)
;;

(* Grad. *)

let grad_engine { Dbg.spf_debug = debug; _; } pm =
  let dm = check_initialized "grad" get_data () in
  debug (sprint_float_matrix "dm" dm);
  M.map2 ( -. ) pm dm
;;

(* Optim. *)

let add_error2 s x y = s +. (x -. y)**2.;;

let cost ctof gd ppm maa =
  let pm = ctof ppm maa in
  0.5 *. (M.fold_left2 add_error2 0. gd pm)
;;

(* ctof = coarse_parameter_to_fine_parameter *)
let optim_engine ctof { Dbg.spf_debug = debug; _; } (ppm, maa_init) =
  let maa_opt = A_ng.array_array_map (fun _ -> 0.) maa_init
  and count = A_ng.array_array_map (fun _ -> ref 0) maa_init in
  let dm = check_initialized "optim" get_data () in
  debug (sprint_float_matrix "dm" dm);
  let dm_sz = M.size dm
  and ppm_sz = M.size ppm in
  debug (sprint_nat_couple "dm_sz" dm_sz);
  debug (sprint_nat_couple "ppm_sz" ppm_sz);
  if dm_sz <> ppm_sz then assert false else
  let add i j zn =
    let maa_opt_i_zn = A_ng.array_array_get maa_opt i zn +. M.get dm i j in
    A_ng.array_array_set maa_opt i zn maa_opt_i_zn;
    incr (A_ng.array_array_get count i zn) in
  M.iter_ij add ppm;
  let os_opt = OS_Convergence in
  let maa_opt =
    A_ng.array_array_map2
      (fun mi ci -> mi /. float_of_int !ci)
      maa_opt count in
  let j_opt = cost ctof dm ppm maa_opt in
  (os_opt, maa_opt, j_opt)
;;

(* Final. *)

let black_float = 0.;;

let final_engine { Dbg.spf_debug = debug; _; } pm =
  let ni = check_initialized "final" get_data_height ()
  and nj = check_initialized "final" get_data_width ()
  and sa = check_initialized "final" get_selection_array () in
  debug (sprint_select_array "sa" sa);
  let fm = unmasking black_float sa pm in
  debug (sprint_float_matrix "fm" fm);
  let img = image_of_float_matrix (get_dim ()) (ni, nj) fm in
  debug (sprint_color_matrix "img" img);
  let fname = get_result_image_file_name () in
  Ppm.save_pixmap fname (M.to_array_array img);
;;
