(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uarray_ng.ml,v 1.8 2017/11/07 23:47:13 fclement Exp $ *)

(* Next generation array. *)

module Nat = Unatural;;
module P = Upervasives;;
module A = Uarray;;
module M = Umatrix;;

let _0_ = Nat._0_;;

let qual = Printf.sprintf "array_%s";;
let invalid_arg = P.invalid_arg_qual qual;;

type index = P.index;;
type offset = index;;
type length = P.length;;

let array_get = A.get;;
let array_unsafe_get = A.unsafe_get;;
let array_set = A.set;;
let array_unsafe_set = A.unsafe_set;;

let array_get_first a =
  let len = A.length a in
  if len = _0_ then invalid_arg "get_first" "empty array" else
  (* Now, we have len > 0.
     Therefore, the following A.unsafe_get will not fail. *)
  A.unsafe_get a _0_
;;

let array_get_last a =
  let len = A.length a in
  if len = _0_ then invalid_arg "get_last" "empty array" else
  (* Now, we have len > 0.
     Therefore, the following Nat.pred and A.unsafe_get will not fail. *)
  A.unsafe_get a (Nat.pred len)
;;

let array_sub = A.sub;;

let array_check_subarray fun_name a ofs len =
  if Nat.add ofs len > A.length a then
    invalid_arg fun_name "wrong subarray specification"
;;

let array_gen_cum_sub_fold_left_i fun_name f accu a ofs len =
  array_check_subarray fun_name a ofs len;
  let r = ref accu in
  let res = A.make (Nat.succ len) accu in
  let rec loop i imax =
    if i >= imax then () else
    let i' = Nat.add ofs i in
    let i = Nat.succ i in
    r := f !r i' (array_unsafe_get a i');
    array_unsafe_set res i !r;
    loop i imax in
  loop _0_ len;
  res
;;

let array_cum_sub_fold_left_i f =
  array_gen_cum_sub_fold_left_i "cum_sub_fold_left_i" f
;;

let array_cum_sub_fold_left f =
  array_gen_cum_sub_fold_left_i "cum_sub_fold_left"
    (fun accu _i x -> f accu x)
;;

let array_cum_fold_left_i f accu a =
  array_gen_cum_sub_fold_left_i "cum_fold_left_i"
    f accu a _0_ (A.length a)
;;

let array_cum_fold_left f accu a =
  array_gen_cum_sub_fold_left_i "cum_fold_left"
    (fun accu _i x -> f accu x) accu a _0_ (A.length a)
;;

let array_for_all_i p a =
  let len = A.length a in
  let rec loop i = i = len || p i (array_get a i) && loop (Nat.succ i) in
  loop _0_
;;

(* Operations on arrays of arrays. *)

let qual = Printf.sprintf "array_array_%s";;
let invalid_arg = P.invalid_arg_qual qual;;

type 'a array_array_t = 'a array array;;

let array_array_get aa i = A.get (A.get aa i);;
let array_array_set aa i = A.set (A.get aa i);;

let array_array_make ni fnj x = A.init ni (fun i -> A.make (fnj i) x);;
let array_array_init ni fnj f = A.init ni (fun i -> A.init (fnj i) (f i));;

let array_array_map f = A.map (A.map f);;
let array_array_map2 f = A.map2 (A.map2 f);;

let array_array_pack aa =
  let lens = A.map A.length aa in
  let a = A.concat (A.to_list aa) in
  (lens, a)
;;

let there_is_zero = A.exists (fun n -> n = _0_);;

let incompatible_lengths len a = A.length a <> len;;

let array_array_unpack lens a =
  if there_is_zero lens then
    invalid_arg "unpack" "zero length" else
  let offsets = array_cum_fold_left Nat.add _0_ lens in
  if incompatible_lengths (array_get_last offsets) a then
    invalid_arg "unpack" "incompatible lengths" else
  let offsets = array_sub offsets _0_ (A.length lens) in
  A.map2 (array_sub a) offsets lens
;;

(* More matrix operations. *)

type line_index = index;;
type column_index = index;;

let matrix_get_line m = array_get (M.to_array_array m);;

let matrix_of_array ni nj a = M.of_array a ni nj;;

let matrix_map3 f m1 m2 m3 =
  M.of_array_array
    (A.map3 (A.map3 f)
       (M.to_array_array m1)
       (M.to_array_array m2)
       (M.to_array_array m3))
;;
