(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uhistory.ml,v 1.28.2.1 2017/11/26 14:37:50 fclement Exp $ *)

let marshal_save_restore () =
  (fun fname (x : 'a) ->
    let oc = open_out_bin fname in
    Marshal.to_channel oc x [ Marshal.Closures; ];
    close_out oc;),
  (fun fname ->
    let ic = open_in_bin fname in
    let x = Marshal.from_channel ic in
    close_in ic;
    (x : 'a))
;;

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L = Ulist;;
module L1 = Ulist1;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;

let _0_ = Nat._0_;;

let string_of_list s_of l =
  let len = L.length l in
  L.fold_left_i
    (fun i accu x ->
      if i = _0_ then accu ^ s_of x else
      if i = Nat.pred len then accu ^ ", and " ^ s_of x else
      accu ^ ", " ^ s_of x)
    "" l
;;

let string_downlift_option f_opt = function
  | "" -> f_opt None
  | s -> f_opt (Some s)
;;

(* Shortcuts for the Basement library. *)

open Uhistory_types;;
open Uhistory_types_print;;

(* History operations. *)

let qual = P.qualify "Uhistory";;
let invalid_arg = P.invalid_arg_qual qual;;

(* Start mode option. *)

let start_mode_option_name = "start mode";;

let default_start_mode = Cold_start;;

let { Ar.get = get_start_mode;
      set = set_start_mode; } =
  Ar.make_aref_single
    start_mode_option_name
    default_start_mode
;;

let start_mode_of_string = function
  | "cold" | "Cold" | "COLD" | "c" | "C" -> Cold_start
  | "warm" | "Warm" | "WARM" | "w" | "W" -> Warm_start
  | s ->
    invalid_arg "start_mode_of_string"
      (Printf.sprintf "Unknown start_mode %S" s)
;;
let set_start_mode_from_string s =
  set_start_mode (start_mode_of_string s)
;;

let start_mode_to_string = function
  | Cold_start -> "cold"
  | Warm_start -> "warm"
;;
let start_mode_to_quoted_string x =
  Printf.sprintf "%S" (start_mode_to_string x)
;;
let print_default_start_mode () =
  start_mode_to_quoted_string (get_start_mode ())
;;
let print_all_start_modes () =
  string_of_list start_mode_to_quoted_string [ Cold_start; Warm_start; ]
;;

let start_mode_lok =
  Ok.long_option_key_of_string start_mode_option_name
and start_mode_sok_st = Ok.Automatic
;;

let start_mode_rco =
  {
   Co.key = start_mode_lok;
   spec = Arg.String set_start_mode_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tSet start mode to <str>.\n\
        \tValid start modes are: %t.\n\
        \tWarm (re)start mode needs option \
          \"-restore-history-file-name\"\n\
        \tto be set (see below).\n\
        \tDefault start mode is %t."
       print_all_start_modes
       print_default_start_mode;
  }
;;

let start_mode_rco_with_sok =
  (start_mode_sok_st, start_mode_rco)
;;

(* History directory name option. *)

let history_directory_name_option_name = "history directory name";;

let default_history_directory_name = ".";;

let { Ar.get = get_history_directory_name;
      set = set_history_directory_name; } =
  Ar.make_aref_single
    history_directory_name_option_name
    default_history_directory_name
;;

let history_directory_name_lok =
  Ok.long_option_key_of_string history_directory_name_option_name
and history_directory_name_sok_st =
  Ok.Short (Ok.of_string "-h-dname")
;;

let history_directory_name_rco =
  {
   Co.key = history_directory_name_lok;
   spec = Arg.String set_history_directory_name;
   doc =
   Printf.sprintf
     "<dname>\n\
      \tSet the directory name in which history files are stored \
        to <dname>.\n\
      \tThis is intended for saved/restored histories, \
        as well as\n\
      \tcheckpointing files (see dedicated options below).\n\
      \tDefault is %S."
    (get_history_directory_name ());
  }
;;

let history_directory_name_rco_with_sok =
  (history_directory_name_sok_st, history_directory_name_rco)
;;

(* Save history file name option. *)

let save_history_file_name_option_name = "save history file name";;

let default_save_history_file_name = None;;

let (get_save_history_file_name, set_save_history_file_name) =
  let { Ar.get; set; } =
    Ar.make_aref_single
      save_history_file_name_option_name
      default_save_history_file_name in
  (get, string_downlift_option set)
;;

let save_history_file_name_lok =
  Ok.long_option_key_of_string save_history_file_name_option_name
and save_history_file_name_sok_st =
  Ok.Short (Ok.of_string "-h-w-fname")
;;

let save_history_file_name_rco =
  {
   Co.key = save_history_file_name_lok;
   spec = Arg.String set_save_history_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet the file name in which history is saved to <fname>.\n\
        \tThe directory name is specified by option \
          \"-history-directory-name\"\n\
        \t(see above).\n\
        \tThis must be used for a further warm (re)start (see above).\n\
        \tAn empty value means no saving.\n\
        \tDefault is no saving.";
  }
;;

let save_history_file_name_rco_with_sok =
  (save_history_file_name_sok_st, save_history_file_name_rco)
;;

(* Restore history file name option. *)

let restore_history_file_name_option_name = "restore history file name";;

let default_restore_history_file_name = "";;

let { Ar.get = get_restore_history_file_name;
      set = set_restore_history_file_name; } =
  Ar.make_aref_unique
    restore_history_file_name_option_name
    default_restore_history_file_name
;;

let restore_history_file_name_lok =
  Ok.long_option_key_of_string
    restore_history_file_name_option_name
and restore_history_file_name_sok_st =
  Ok.Short (Ok.of_string "-h-r-fname")
;;

let restore_history_file_name_rco =
  {
   Co.key = restore_history_file_name_lok;
   spec = Arg.String set_restore_history_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet the file name from which history is restored to <fname>.\n\
        \tThe directory name is specified by option \
          \"-history-directory-name\"\n\
        \t(see above).\n\
        \tUseful (and mandatory) for warm (re)start only (see above).";
  }
;;

let restore_history_file_name_rco_with_sok =
  (restore_history_file_name_sok_st, restore_history_file_name_rco)
;;

(* History step mode option. *)

let history_step_more_option_name = "history step mode";;

let default_history_step_mode = Last;;

let { Ar.get = get_history_step_mode;
      set = set_history_step_mode; } =
  Ar.make_aref_single
    history_step_more_option_name
    default_history_step_mode
;;

let history_step_mode_of_string = function
  | "last" | "Last" | "LAST" | "l" | "L" -> Last
  | "full" | "Full" | "FULL" | "f" | "F" -> Full
  | s ->
    invalid_arg "history_step_mode_of_string"
      (Printf.sprintf "Unknown history step mode %S" s)
;;
let set_history_step_mode_from_string s =
  set_history_step_mode (history_step_mode_of_string s)
;;

let history_step_mode_to_string = function
  | Last -> "last"
  | Full -> "full"
;;
let history_step_mode_to_quoted_string x =
  Printf.sprintf "%S" (history_step_mode_to_string x)
;;
let print_default_history_step_mode () =
  history_step_mode_to_quoted_string (get_history_step_mode ())
;;
let print_all_history_step_modes () =
  string_of_list history_step_mode_to_quoted_string [ Last; Full; ]
;;

let history_step_mode_lok =
  Ok.long_option_key_of_string history_step_more_option_name
and history_step_mode_sok_st =
  Ok.Short (Ok.of_string "-h-step")
;;

let history_step_mode_rco =
  {
   Co.key = history_step_mode_lok;
   spec = Arg.String set_history_step_mode_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tSet history step mode to <str>.\n\
        \tValid history step modes are: %t.\n\
        \tDefault history step mode is %t."
       print_all_history_step_modes
       print_default_history_step_mode;
  }
;;

let history_step_mode_rco_with_sok =
  (history_step_mode_sok_st, history_step_mode_rco)
;;

(* Checkpointing option. *)

let checkpointing_option_name = "checkpointing";;

let default_checkpointing = None;;

let (get_checkpointing, set_checkpointing) =
  let { Ar.get; set; } =
    Ar.make_aref_single
      checkpointing_option_name
      default_checkpointing in
  (get, string_downlift_option set)
;;

let checkpointing_lok =
  Ok.long_option_key_of_string checkpointing_option_name
and checkpointing_sok_st =
  Ok.Short (Ok.of_string "-ckp")
;;

let checkpointing_rco =
  {
   Co.key = checkpointing_lok;
   spec = Arg.String set_checkpointing;
   doc =
     Printf.sprintf
       "<str>\n\
        \tSet the basename of checkpointing files to <str>.\n\
        \tThe directory name is specified by option \
          \"-history-directory-name\"\n\
        \t(see above).\n\
        \tAn empty value means no checkpointing.\n\
        \tDefault is no checkpointing.";
  }
;;

let checkpointing_rco_with_sok =
  (checkpointing_sok_st, checkpointing_rco)
;;

let get_history_mode () = (get_history_step_mode (), get_checkpointing ());;

(* Command-line options. *)

let phase_start_mode = Co.Phase "start_mode";;

let add_history_options () =
  Co.application_add_all_rco_with_sk
    [
     start_mode_rco_with_sok;
     history_directory_name_rco_with_sok;
     save_history_file_name_rco_with_sok;
     restore_history_file_name_rco_with_sok;
     history_step_mode_rco_with_sok;
     checkpointing_rco_with_sok;
    ];
  Co.phase_add_all
    phase_start_mode
    [ start_mode_lok;
      restore_history_file_name_lok;
      history_directory_name_lok; ];
;;

let add_dname f fname =
  let dname = get_history_directory_name () in
  f (Filename.concat dname fname)
;;

module Make =
  functor (Config : Uhistory_spec.CONFIGURATION) ->
  functor (St : Uhistory_spec.STATE) ->
struct
  type config = Config.t;;
  type constants = St.constants;;
  type state = St.t;;

  type t = (config, constants, state) Uhistory_types.t;;

  (* Late evaluation to insert these options after those of the algorithm. *)
  add_history_options ();;

  let (save_history, restore_history) =
    let (save, restore) = marshal_save_restore () in
    (add_dname save, add_dname restore)
  ;;

  let head_state { h_list = ({ state = sti; _; }, _); _; } = sti;;
  let head_state_number { h_list = ({ state_number = i; _; }, _); _; } = i;;

  let make_trivial_history_list i sti =
    ({ state_number = i; state = sti; }, [])
  ;;

  let last_history_step i sti h =
    { h with h_list = make_trivial_history_list i sti; }
  ;;

  let full_history_step i sti ({ h_list = hlim1; _; } as h) =
    { h with h_list = L1.cons { state_number = i; state = sti; } hlim1; }
  ;;

  let filename_of_basename bname h =
    let i = head_state_number h in
    Printf.sprintf "%s_%d.obj" bname (Nat.to_int i)
  ;;

  let checkpoint_aux ckp h =
    match ckp with
    | None -> ""
    | Some bname ->
      let fname = filename_of_basename bname h in
      save_history fname h;
      fname
  ;;

  let cold_ckp = function
    | None -> Some "cold"
    | _ as ckp -> ckp
  ;;

  let cold_checkpoint ckp = checkpoint_aux (cold_ckp ckp)
  and checkpoint ckp h = ignore (checkpoint_aux ckp h)
  ;;

  let make_history_step i sti ({ h_mode = (hs_mode, ckp); _; } as him1) =
    let hi =
      match hs_mode with
      | Last -> last_history_step i sti him1
      | Full -> full_history_step i sti him1 in
    checkpoint ckp hi;
    hi
  ;;

  let same_head_state h1 h2 =
    St.same (head_state h1) (head_state h2)
  ;;

  (* Printer shortcuts. *)

  let printer ctx =
    printer_t Config.printer St.printer_constants St.printer ctx
  ;;

  let print ppf = printer Og_precedence.Ck_toplevel ppf;;

  let printer_head_h_state ctx ppf h =
    printer_h_state St.printer ctx ppf (L1.hd h.h_list)
  ;;

  let print_head_h_state ppf =
    printer_head_h_state Og_precedence.Ck_toplevel ppf
  ;;

  let initialization () =
    Co.parse_phase phase_start_mode;
    let rec loop = function
      | Cold_start ->
        Co.parse ();
        Config.cold_set ();
        let st0 = St.init () in
        let (_, ckp as h_mode) = get_history_mode () in
        let h0 =
          {
           configuration = Config.get ();
           constants = St.init_constants st0;
           h_mode;
           h_list = make_trivial_history_list _0_ st0;
          } in
        let fname = cold_checkpoint ckp h0 in
        set_restore_history_file_name fname;
        loop Warm_start;
      | Warm_start ->
        let h0 = restore_history (get_restore_history_file_name ()) in
        Co.parse ();
        Config.warm_set h0.configuration;
        { h0 with configuration = Config.get (); } in
    loop (get_start_mode ())
  ;;

  let finalization hn =
    let h_state_n = head_state hn in
    St.final h_state_n;
    match get_save_history_file_name () with
    | None -> ()
    | Some fname -> save_history fname hn
  ;;
end
;;
