(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_mask.ml,v 1.9 2017/11/07 20:57:40 fclement Exp $ *)

(* Matrix masks. *)

module P = Upervasives;;
module Nat = Unatural;;
module L = Ulist;;
module M = Umatrix;;

let _0_ = Nat._0_;;

let qual = P.qualify "Umatrix_mask";;
let failwith = P.failwith_qual qual;;

type length = P.length;;

type selection_matrix = P.select M.t;;

type size = P.length * P.length;;

type nat_couple = Nat.t * Nat.t;;

type mask =
  {
   size : size;
   body : mask_body;
  }
and t = mask
and mask_body =
  | Empty
  | Cell of cell
  | Range of range
  | Rectangle of corners
  | Union of mask_body list
and corners =
  {
   upper_left : cell;
   lower_right : cell;
  }
and range =
  {
   left : cell;
   right : cell;
  }
and cell = nat_couple
;;

let cell_iter f c = f c;;

let range_iter_j f i j = cell_iter f (i, j);;

let unsafe_range_iter_i f jl jr i = Nat.for_loop jl jr (range_iter_j f i);;

let range_iter_i f nj (il, jl) (ir, jr) i =
 let njm1 = Nat.pred nj in
 let jmin = if i = il then jl else _0_ in
 let jmax = if i = ir then min jr njm1 else njm1 in
 unsafe_range_iter_i f jmin jmax i
;;

let unsafe_range_iter f { left = (il, jl); right = (ir, jr); } =
  Nat.for_loop il ir (unsafe_range_iter_i f jl jr)
;;

let range_iter f (ni, nj) { left = (il, _ as l); right = (ir, _ as r); } =
  let imax = min ir (Nat.pred ni) in
  Nat.for_loop il imax (range_iter_i f nj l r)
;;

let rectangle_iter_i f jl jr i =
  let rg = { left = (i, jl); right = (i, jr); } in
  unsafe_range_iter f rg
;;

let i_bounds { upper_left = (iul, _); lower_right = (ilr, _); } = (iul, ilr)
and j_bounds { upper_left = (_, jul); lower_right = (_, jlr); } = (jul, jlr)
;;

let rectangle_iter f (ni, nj) cs =
  let (iul, ilr) = i_bounds cs
  and (jul, jlr) = j_bounds cs in
  let imax = min ilr (Nat.pred ni)
  and jmax = min jlr (Nat.pred nj) in
  Nat.for_loop iul imax (rectangle_iter_i f jul jmax)
;;

let rec union_iter f sz mbl = L.iter (mask_body_iter f sz) mbl

and mask_body_iter f sz = function
  | Empty -> ()
  | Cell c -> cell_iter f c
  | Range rg -> range_iter f sz rg
  | Rectangle cs -> rectangle_iter f sz cs
  | Union mbl -> union_iter f sz mbl
;;

let mask_iter f { size = sz; body = mb; } = mask_body_iter f sz mb;;

let failwith_set_cell s i =
  failwith "set_cell"
    (Printf.sprintf "out of bound %s index %d" s (Nat.to_int i))
;;

let forget_cell sm (ni, nj) (i, j) =
  if ni <= i then failwith_set_cell "line" i else
  if nj <= j then failwith_set_cell "column" j else
  M.set sm i j P.Forget
;;

let set_selection_matrix sm ({ size = sz; _; } as mask) =
  mask_iter (forget_cell sm sz) mask
;;

let selection_matrix ({ size = (ni, nj); _; } as mask) =
  let sm = M.make ni nj P.Keep in
  set_selection_matrix sm mask;
  sm
;;

let selection_length =
  M.fold_left
    (fun accu -> function
      | P.Forget -> accu
      | P.Keep -> Nat.succ accu)
    _0_
;;
