(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_structural_mapping.ml,v 1.18 2017/11/07 20:57:40 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module A = Uarray;;
module Dg = Udynamic_graph;;
module Ddg = Udelayed_dgraph;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;

(* Shortcuts for the Mapping library. *)

module Iidx = Uindex_index;;
module Midx = Umatrix_index;;
module Iuniv = Uindex_universe;;
module Muniv = Umatrix_universe;;
module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;

module Set = Mset;;
module Map = Mg;;

(* Matrix structural mapping operations. *)

type 'a mapping = 'a Map.t
and 'a sub_mapping = 'a Map.t
;;

type 'a partition = 'a sub_mapping array;;

type block_number = Ddg.arrow_number;;

type ('a, 'b) block = ('a sub_mapping, 'b) Ddg.delayed_arrow;;

type ('a, 'b, 'c) t = ('a sub_mapping, 'b, 'c) Ddg.t;;

let make f = Ddg.make f _1_;;

let same = Ddg.same;;

let equal_images equal_image =
  Ddg.equal_values (Map.equal_images equal_image)
;;

(* Access functions. *)

let get = Ddg.get;;
let set = Ddg.set;;
let set_next = Ddg.set_next;;

let last_block_number msm =
  Nat.pred (Dg.next_unset msm.Ddg.graph)
;;

(* Converters. *)

let restriction
    ({ Map.domain = dom; _} as map) { Map.domain = sub_dom; _} =
  let ss = Set.subset_of_full dom sub_dom in
  Map.restriction map ss
;;

let partition_of_mapping pp map = A.map (restriction map) pp;;

let set_zn_pt geom i _idx uidx =
  A.set geom (Midx.to_nat uidx) i
;;

let set_zn geom i { Map.domain = dom; _; } =
  Set.iter (set_zn_pt geom i) dom
;;

let partition_to_geometry pp =
  let nbz = A.length pp in
  if nbz = _0_ then assert false else
  let { Map.domain = dom0; _; } = A_ng.array_get_first pp in
  let nbpt = Muniv.cardinal dom0.Set.universe in
  let geom = A.make nbpt _0_ in
  A.iter_i (set_zn geom) pp;
  geom
;;

let partition_to_values pp = A.map Map.get_constant pp;;

let partition_to_geometry_and_values pp =
  (partition_to_geometry pp, partition_to_values pp)
;;

let one_zone_partition sz m0 =
  let dom0 = Set.make sz in
  [| Map.constant dom0 m0; |]
;;

let full_domain_of_msm msm =
  let { Dg.value = map; _; } = get msm _0_ in
  let u = map.Map.domain.Set.universe in
  Set.full u
;;

let msm_choice msm =
  let { Dg.value = map; _; } = get msm _0_ in
  match map.Map.images with
  | Map.Constant y -> y
  | Map.Images ims -> A_ng.array_get_first ims
;;

let images_of_msm msm dom =
  let y0 = msm_choice msm in
  let ims = A.make (Set.cardinal dom) y0 in
  let f_item _i { Dg.value = item_map; _; } =
    match item_map with
    | { Map.domain = item_dom;
        images = Map.Constant item_y; } ->
        let f _ item_uidx =
          let idx = Iuniv.index dom.Set.uindices item_uidx in
          let i = Iidx.to_nat idx in
          A.set ims i item_y in
        Set.iter f item_dom
    | { Map.domain = item_dom;
        images = Map.Images item_ims; } ->
        let f item_idx item_uidx =
          let idx = Iuniv.index dom.Set.uindices item_uidx in
          let i = Iidx.to_nat idx in
          let item_i = Iidx.to_nat item_idx in
          A.set ims i (A.get item_ims item_i) in
        Set.iter f item_dom in
  Dg.iter f_item msm.Ddg.graph;
  ims
;;

let to_mapping msm =
  let dom = full_domain_of_msm msm in
  let ims = images_of_msm msm dom in
  Map.from_array dom ims
;;

let to_array = Ddg.to_array;;

let of_partition = Ddg.of_values
and to_partition { Ddg.graph = dg; _; } = Dg.to_values dg
;;

(* Iterators. *)

let iter = Ddg.iter
and fold_left = Ddg.fold_left
and partition_iter = A.iter_i
and iter_to_do = Ddg.iter_to_do
and iter_value = Ddg.iter_value
and fold_left_value = Ddg.fold_left_value
;;
