(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: refinement.ml,v 1.76.2.1 2017/11/17 09:18:56 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L = Ulist;;
module Ddg = Udelayed_dgraph;;

(* Shortcuts for the Basement library. *)

open Uhistory_types;;

(* Shortcuts for the Mapping library. *)

module Msm = Umatrix_structural_mapping;;

module Smap = Msm;;

(* Shortcuts for the Refinement library. *)

module Config = Configuration;;

open Refinement_types;;
open Refinement_types_print;;
open Optimization_status_types;;
open Stopping_criterion_types;;
open Constants_types;;

(* Refinement indicator algorithm. *)

let qual = P.qualify "Refinement";;
let print_debug fname = Config.print_debug (qual fname);;

let print_Float = Og_printer.printer_Float Og_precedence.Ck_toplevel;;

module Make =
  functor (Scal : Scalar_spec.SCALAR) ->
  functor (Comps : Components_spec.COMPONENTS) ->
struct
  module D = Display.Make (Scal) (Comps);;
  module Bb = D.Bb;;
  module St = D.St;;
  module H = D.H;;
  module Stt = Bb.Stt;;

  let is_convergence { Stt.stopping_criterion = sc; _; } =
    sc = OS_Convergence
  ;;

  let selected_refinement_with_smallest_cost srl =
    let srl = L.filter is_convergence srl in
    let compare { Stt.cost = j1; _; } { Stt.cost = j2; _; } =
      Pervasives.compare j1 j2 in
    let srl = L.sort compare srl in
    match srl with
    | [] -> raise Not_found
    | x :: _ -> x
  ;;

  let print_gradients = Comps.print (printer_gradient Scal.printer);;

  let ppf = Config.get_formatter ();;
  let scc = Config.get_stopping_criterion_config ();;

  let display_msg = Format.fprintf ppf "@[%s@]@.";;

  (* Iteration. *)

  (* Step 1. *)
  (* state -> gradient *)
  let compute_fine_gradient = Bb.grad;;

  (* Step 2a. *)
  (* state -> gradients -> forced_zonations *)
  let compute_scalar_first_order_indicators { Stt.zonations = zs; _; } gs =
    (* The following map could be parallel. *)
    let zs = Comps.map2 Ddg.do_it gs zs in
    Comps.map Smap.to_array zs
  ;;

  (* Step 2b. *)
  (* state -> gradients -> forced_zonations *)
  let compute_candidate_refinements st gs =
    let fzs = compute_scalar_first_order_indicators st gs in
    let vector_first_order_indicators =
      match Config.get_refinement_strategy () with
      | Overall_best ->
        St.vector_first_order_indicators_overall_best
      | Best_in_a_family (_, _) ->
        St.vector_first_order_indicators_best_in_a_family in
    match Config.get_dimension_management () with
    | Best_component_only -> fzs
    | Vector -> vector_first_order_indicators fzs
  ;;

  (* Step 3. *)
  (* forced_zonations -> partitions list *)
  let select_best_candidate_refinements fzs =
    let (ratio, nb) =
      Config.get_candidate_refinements_selection_criterion () in
    let foi_max = St.max_first_order_indicator fzs in
    let threshold = Scal.( *. ) (Scal.of_float ratio) foi_max in
    let select foi = foi >= threshold in
    St.select_n_best_cc fzs select nb
  ;;

  (* Step 4. *)
  (* state -> partitions list -> selected_refinement list *)
  let compute_exact_indicator_on_selected_refinements st ppsl =
    (* The following map should be parallel. *)
    L.map (Bb.optim st) ppsl
  ;;

  (* Step 5. *)
  (* state -> partitions list -> selected_refinement list -> state *)
  let pick_best_selected_refinement st ppsl srl =
    let (st, sc) =
      match (ppsl, srl) with
      | ([], []) -> (st, SC_No_more_possible_refinements)
      | ([], _ :: _) | (_ :: _, []) -> assert false
      | (_, _ :: _) ->
        if St.all_equal_coarse_parameters ppsl srl then
          (st, SC_Stagnation) else
        try
          let srl_opt = selected_refinement_with_smallest_cost srl in
          ({ srl_opt with Stt.stopping_criterion = SC_Continue; },
           SC_Continue)
        with
        | Not_found -> (st, SC_No_convergent_optimizations) in
    { st with Stt.stopping_criterion = sc; }
  ;;

  let test_max_iteration i =
    i >= scc.max_iteration
  ;;

  let test_min_relative_gradients_norm ng0 ngs_i =
    ngs_i /. ng0 <= scc.min_relative_gradients_norm
  ;;

  (* (state -> partitions list -> selected_refinement list) ->
     history -> history *)
  let iteration optims sth_i =
    let i = H.head_state_number sth_i in
    let ip1 = Nat.succ i in
    let st_i = H.head_state sth_i in
    let st_ip1 =
      if test_max_iteration i then
        begin
          display_msg "STOP: max_iteration";
          { st_i with Stt.stopping_criterion = SC_Too_many_iterations; }
        end else
      let constants = sth_i.constants in
      let ng0 = constants.initial_gradients_norm in
      let (gs_i, ngs_i) = compute_fine_gradient st_i in
      print_debug "iteration" "gs_i" print_gradients ppf gs_i;
      print_debug "iteration" "ngs_i" print_Float ppf ngs_i;
      D.display_grad st_i gs_i;
      if test_min_relative_gradients_norm ng0 ngs_i then
        begin
          display_msg "CONV: min_relative_gradients_norm";
          { st_i with Stt.stopping_criterion = SC_Convergence; }
        end else
      let fzs_i = compute_candidate_refinements st_i gs_i in
      let ppsl_i = select_best_candidate_refinements fzs_i in
      let srl_i = optims st_i ppsl_i in
      pick_best_selected_refinement st_i ppsl_i srl_i in
    H.make_history_step ip1 st_ip1 sth_i
  ;;

  (* (state -> partitions list -> selected_refinement list) ->
     history -> history *)
  let rec iterations optims sth_i =
    print_debug "iteration" "sth_i" H.print_head_h_state ppf sth_i;
    D.display_param sth_i;
    let sth_ip1 = iteration optims sth_i in
    let { Stt.stopping_criterion = sc_ip1; _; } = H.head_state sth_ip1 in
    match sc_ip1 with
    | SC_Continue -> iterations optims sth_ip1
    | SC_Convergence
    | SC_No_convergent_optimizations
    | SC_No_more_possible_refinements
    | SC_Stagnation
    | SC_Too_many_iterations -> sth_ip1
  ;;

  (* Algorithm. *)

  (* (state -> partitions list -> selected_refinement list) -> unit *)
  let algorithm optims =
    let sth_1 = H.h_init in
    print_debug "algorithm" "sth_1" H.print_head_h_state ppf sth_1;
    D.init sth_1;
    let sth_n = iterations optims sth_1 in
    H.finalization sth_n;
  ;;
end
;;
