(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: refinement_types.mli,v 1.41 2017/11/07 20:57:41 fclement Exp $ *)

(** {3 Types for the refinement indicator algorithm} *)

type component_number = Unatural.t;;

(** {6 External representation of parameters}
 Extensional types for parameter refinement. *)

type point = Unatural.t * Unatural.t
(** A [point] denotes a matrix entry. *)
;;

type 'a parameter = 'a Umatrix_structural_mapping.mapping
and 'a sub_parameter = 'a Umatrix_structural_mapping.sub_mapping
(** An ['a parameter] is the graph of the function that associates each
 point of the domain of a parameter with ['a] values. *)
;;

type 'a gradient = 'a parameter
(** The gradient of a scalar function with respect to a parameter has the same
 representation as the parameter itself. *)
;;

(** {6 Internal representation of parameters}
 Structural types for parameter refinement. *)

type cost = float
and 'a first_order_indicator = 'a (* 'a is the scalar type of parameters. *)
and exact_indicator = cost
and gradients_norm = float
and initial_gradients_norm = gradients_norm
(** The type for first order and exact indicators. *)
;;

type 'a candidate_refinement =
  {
   refinement : Umatrix_nonempty_set.dichotomy;
   first_order_indicator : 'a first_order_indicator;
  }
(** An ['a candidate_refinement] is the association of a dichotomy of the
 support of a sub-part of a parameter with the corresponding first order
 indicator. *)
;;

type 'a zone =
  ('a, 'a candidate_refinement list) Umatrix_structural_mapping.block
(** An ['a zone] is the association of a sub-part of a parameter with a list of
 candidate refinements. The evaluation of the list may be delayed. *)
;;

type zone_number = Umatrix_structural_mapping.block_number
(** Zones are numbered. *)
;;

type bounds = { ll : point; ur : point; }
(** Bounds of zones are represented by lower left and upper right corners
 of the smallest rectangle containing the support of the zone. *)
;;

type 'a zonation =
  ('a, 'a candidate_refinement list, 'a gradient) Umatrix_structural_mapping.t
(** An ['a zonation] is a collection of zones.
 The intention is to only consider zones with disjoint supports.
 ['a] will always be [scalar].
 The evaluation of the lists of candidate refinements may be delayed. *)
;;

type 'a forced_zone =
  ('a sub_parameter, 'a candidate_refinement list) Udynamic_graph.arrow
;;

type 'a forced_zonation = 'a forced_zone array;;

type 'a partition = 'a sub_parameter array;;

(** {6 Refinement algorithm} *)

type command_line =  string;;
type external_program_command = command_line;;

type dimension = Unatural.t;;

type dimension_management =
  | Vector              (* = (Best_refinement, All_components) *)
  | Best_component_only (* = (Best_refinement, Best_component) *)
;;

type number_of_cuts = Unatural.t;;

type basic_refinement_family =
  | Horizontal_cut
  | Vertical_cut
;;

type refinement_strategy =
  | Best_in_a_family of number_of_cuts * basic_refinement_family Ulist1.list1
  | Overall_best
;;

type candidate_refinements_selection_criterion = float * Unatural.t;;

type stopping_criterion_config =
  {
   min_relative_gradients_norm : float;
   min_relative_best_scalar_indicator : float;
   min_relative_scalar_cost_decrease : float;
   max_iteration : Uhistory_types.state_number;
  }
;;

type graphics_mode =
  | No_graphics        (* = None *)
  | Parameter_graphics (* = Some (Param_quantity, All_channels) *)
  | Full_graphics      (* = Some (All_quantities, All_channels) *)
;;

type output_status =
  | Max_iteration of Uhistory_types.state_number
  | Min_relative_best_scalar_indicator of float
  | Min_relative_scalar_cost_decrease of float
  | No_more_refinements
;;
