(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: wrapper.ml,v 1.29 2017/11/07 20:57:41 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module A = Uarray;;
module M = Umatrix;;
module Dbg = Udebug;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
and _4_ = Nat._4_
and _5_ = Nat._5_
and _6_ = Nat._6_
;;

(* Shortcuts for the Pio library. *)

module Pio_tv = Typed_value;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;

open Misc_sprint;;

(* Shortcuts for the Refinement library. *)

module Os = Optimization_status;;

open Optimization_status_types_print;;
open Wire_types_print;;

(* Wrapper operations. *)

let qual = P.qualify "Wrapper";;
let failwith = P.failwith_qual qual;;

type ('a, 'b) wrapper_maker =
  Pio.pf -> P.function_name -> ('a, 'b) Pio.task_wrapper
;;

(* Init. *)

let mk_init_inj { Dbg.spf_report = report; spf_debug = debug; _; } dim =
  report "sending one value";
  report "sending dimension";
  debug (format_sprint printer_dimension "dim" dim);
  [| Pio_tv.of_int (Nat.to_int dim); |]
;;

let mk_init_proj { Dbg.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = A.length vals in
  if nargs <> _6_ then
    failwith "init_proj" "initialization should return six values" else
  report "six values received";
  let (ni, nj) = Pio_tv.to_int_couple (A.get vals _0_) in
  let sz = (Nat.of_int ni, Nat.of_int nj) in
  report "domain size extracted";
  debug (format_sprint printer_size "sz" sz);
  let mpm = Pio_tv.to_c_int_matrix (A.get vals _1_) in
  report "initial masked partition extracted";
  debug (format_sprint printer_masked_partition "mpm" mpm);
  let sos = Pio_tv.to_string (A.get vals _2_) in
  let os = Os.scan sos in
  report "initial optimization status extracted";
  debug (format_sprint printer_optimization_status "os" os);
  let lens = Pio_tv.to_int_vector (A.get vals _3_) in
  let lens = A.map Nat.of_int lens in
  report "initial coarse parameter lengths extracted";
  let a = Pio_tv.to_float_vector (A.get vals _4_) in
  let maa = A_ng.array_array_unpack lens a in
  report "initial coarse parameter extracted";
  debug (format_sprint printer_coarse_parameter "maa" maa);
  let j = Pio_tv.to_float (A.get vals _5_) in
  report "initial cost extracted";
  debug (format_sprint printer_cost "j" j);
  (sz, mpm, (os, maa, j))
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

(* Grad. *)

let mk_grad_inj { Dbg.spf_report = report; spf_debug = debug; _; } pm =
  report "sending one value";
  report "sending fine parameter";
  debug (format_sprint printer_fine_parameter "pm" pm);
  [| Pio_tv.of_c_float_matrix pm; |]
;;

let mk_grad_proj { Dbg.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = A.length vals in
  if nargs <> _1_ then
    failwith "grad_proj" "gradient should return one value" else
  report "one value received";
  let gm = Pio_tv.to_c_float_matrix (A.get vals _0_) in
  report "gradient matrix extracted";
  debug (format_sprint printer_fine_gradient "gm" gm);
  gm
;;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

(* Optim. *)

let mk_optim_inj
    { Dbg.spf_report = report; spf_debug = debug; _; } (ppm, maa_init) =
  report "sending three values";
  report "sending partition";
  debug (format_sprint printer_partition "ppm" ppm);
  let ppm = M.map Nat.to_int ppm in
  report "sending initial coarse parameter";
  debug (format_sprint printer_coarse_parameter "maa_init" maa_init);
  let (lens_init, a_init) = A_ng.array_array_pack maa_init in
  let lens_init = A.map Nat.to_int lens_init in
  [| Pio_tv.of_c_int_matrix ppm;
     Pio_tv.of_int_vector lens_init;
     Pio_tv.of_float_vector a_init; |]
;;

let mk_optim_proj { Dbg.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = A.length vals in
  if nargs <> _4_ then
    failwith "optim_proj" "optimization should return four values" else
  report "four values received";
  let sos_opt = Pio_tv.to_string (A.get vals _0_) in
  let os_opt = Os.scan sos_opt in
  report "optimization status extracted";
  debug (format_sprint printer_optimization_status "os_opt" os_opt);
  let lens_opt = Pio_tv.to_int_vector (A.get vals _1_) in
  let lens_opt = A.map Nat.of_int lens_opt in
  report "optimal coarse parameter lengths extracted";
  let a_opt = Pio_tv.to_float_vector (A.get vals _2_) in
  let maa_opt = A_ng.array_array_unpack lens_opt a_opt in
  report "optimal coarse parameter extracted";
  debug (format_sprint printer_coarse_parameter "maa_opt" maa_opt);
  let j_opt = Pio_tv.to_float (A.get vals _3_) in
  report "optimal cost extracted";
  debug (format_sprint printer_cost "j_opt" j_opt);
  (os_opt, maa_opt, j_opt)
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

(* Final. *)

let mk_final_inj { Dbg.spf_report = report; spf_debug = debug; _; } pm =
  report "sending one value";
  report "sending fine parameter";
  debug (format_sprint printer_fine_parameter "pm" pm);
  [| Pio_tv.of_c_float_matrix pm; |]
;;

let mk_final_proj { Dbg.spf_report = report; spf_debug = debug; _; } vals =
  let nargs = A.length vals in
  if nargs <> _1_ then
    failwith "final_proj" "finalization should return one value" else
  report "one value received";
  let msg = Pio_tv.to_string (A.get vals _0_) in
  debug (format_sprint Og_printer.printer_String "msg" msg);
  if msg <> "OK, bye!" then
    failwith "final_proj" "finalization should return \"OK, bye!\"" else
  report "farewell message extracted";
;;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;
