(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: zonation.ml,v 1.33 2017/11/07 20:57:41 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module A = Uarray;;
module Dg = Udynamic_graph;;

let _0_ = Nat._0_;;

(* Shortcuts for the Mapping library. *)

module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;
module Msm = Umatrix_structural_mapping;;

module Set = Mset;;
module Map = Mg;;
module Smap = Msm;;

(* Shortcuts for the Refinement library. *)

open Refinement_types;;

(* Zonation operations. *)

let same = Smap.same
and last_zone_number = Smap.last_block_number
;;

(* Partitions. *)

let partition_unique_domains pp0 cn pp =
  if cn = _0_ then pp0 else A.map2 Map.unique_domain pp0 pp
;;

let partition_to_geometry_and_values =
  Smap.partition_to_geometry_and_values
(*and partition_of_geometry_and_values =
  Smap.partition_of_geometry_and_values*)
and one_zone_partition = Smap.one_zone_partition
;;

(* Converters. *)

let partition_of_gradient = Smap.partition_of_mapping
and to_parameter = Smap.to_mapping
and of_partition = Smap.of_partition
and to_partition = Smap.to_partition
;;

let equal_coarse_parameter pp z =
  A.for_all2 (Map.equal_images Pervasives.( = )) pp (to_partition z)
;;

(* Iterators. *)

let iter = Smap.iter
and fold_left = Smap.fold_left
and partition_iter = Smap.partition_iter
and iter_to_do = Smap.iter_to_do
and iter_value = Smap.iter_value
and fold_left_value = Smap.fold_left_value
;;

let force = Smap.to_array;;

(* Refinements. *)

let cartesian_cuts ijmin ijmax dom p nct =
  (* nij = size of the box *)
  let nij = Nat.succ (Nat.sub ijmax ijmin) in
  (* nct is the maximum number of cuts,
     therefore nct + 1 is the maximum number of intervals.
     The value -1 means all possible cuts. *)
  let nint = if nct = _0_ then nij else Nat.succ nct in
  let ijl = Nat.psplit nij nint in
  let rec loop accu = function
    | ([], _) -> assert false
    | (k :: kl, ijc) when ijc = ijmax ->
        if kl = [] then accu else
        loop accu (kl, Nat.sub ijc k)
    | (k :: kl, ijc) ->
        let accu =
          match Set.dichotomy (p ijc) dom with
          | None -> accu
          | Some dicho -> dicho :: accu in
        if kl = [] then accu else
        loop accu (kl, Nat.sub ijc k) in
  loop [] (ijl, ijmax)
;;

let horizontal_cuts { ll = (imin, _); ur = (imax, _); } dom nct =
  let pi ic idx _uidx =
    let (i, _) = Set.element dom idx in
    i <= ic in
  cartesian_cuts imin imax dom pi nct
;;

let vertical_cuts { ll = (_, jmin); ur = (_, jmax); } dom nct =
  let pj jc idx _uidx =
    let (_, j) = Set.element dom idx in
    j <= jc in
  cartesian_cuts jmin jmax dom pj nct
;;

let dichotomies_from_family bds dom nct = function
  | Horizontal_cut -> horizontal_cuts bds dom nct
  | Vertical_cut -> vertical_cuts bds dom nct
;;

let same_refinement acr cr = acr.refinement == cr.refinement;;

let same_zone_geometry z1 z2 =
  Parameter.same_geometry z1.Dg.value z2.Dg.value
;;
