(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: worker.ml,v 1.28 2017/11/07 20:57:42 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L1 = Ulist1;;
module A = Uarray;;
module M = Umatrix;;
module Dbg = Udebug;;

let _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
;;

(* Shortcuts for the Pio library. *)

module Pio_tv = Typed_value;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;

open Misc_sprint;;

(* Shortcuts for the Refinement library. *)

module Os = Optimization_status;;

let sprint_optimization_status = Os.sprint;;

(* Generic part of Ocaml workers for the inversion platform. *)

open Worker_types;;

let qual = P.qualify "Worker";;
let invalid_arg = P.invalid_arg_qual qual;;
let failwith = P.failwith_qual qual;;

(* Tasks. *)

let check_same_dimensions fun_name msg dim dim' =
  if dim' <> dim then invalid_arg fun_name msg
;;

let check_valid_dimension fun_name dim dim' =
  let msg = Printf.sprintf "dimension must be %d" (Nat.to_int dim) in
  check_same_dimensions fun_name msg dim dim'
;;

let check_valid_zone_numbers fun_name ppm maa =
  let valid_zone_numbers =
    A_ng.array_for_all_i
      (fun i ->
        A.for_all
          (fun pp_ij -> pp_ij < A.length (A_ng.array_get maa i)))
      (M.to_array_array ppm) in
  if not valid_zone_numbers then
    invalid_arg fun_name "zone numbers in partition are not all valid"
;;

let coarse_parameter_to_fine_parameter ppm maa =
  M.map_ij (fun i _j pp_ij -> A_ng.array_array_get maa i pp_ij) ppm
;;

let make_init
    init_engine optim ({ Dbg.spf_debug = debug; _; } as spf) dim =
  debug (sprint_nat "dim" dim);
  let (sz, mpm1, ppm1, maa1_init) = init_engine spf dim in
  debug (sprint_nat_couple "sz" sz);
  debug (sprint_int_matrix "mpm1" mpm1);
  let (sc1, maa1, j1 as optim_out1) = optim spf (ppm1, maa1_init) in
  debug (sprint_optimization_status "sc1" sc1);
  debug (sprint_float_array_array "maa1" maa1);
  debug (sprint_float "j1" j1);
  (sz, mpm1, optim_out1)
;;

let make_grad grad_engine ({ Dbg.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  let gm = grad_engine spf pm in
  debug (sprint_float_matrix "gm" gm);
  gm
;;

let make_optim
    optim_engine ({ Dbg.spf_debug = debug; _; } as spf) (ppm, maa_init) =
  debug (sprint_nat_matrix "ppm" ppm);
  debug (sprint_float_array_array "maa_init" maa_init);
  let dim = A.length maa_init in
  let dim' = M.dim_i ppm in
  let msg =
    "arguments partition and initial coarse parameter \
     have incompatible dimensions" in
  check_same_dimensions "optim" msg dim dim';
  check_valid_zone_numbers "optim" ppm maa_init;
  let ((sc_opt, maa_opt, j_opt) as optim_out_opt) =
    optim_engine spf (ppm, maa_init) in
  debug (sprint_optimization_status "sc_opt" sc_opt);
  debug (sprint_float_array_array "maa_opt" maa_opt);
  debug (sprint_float "j_opt" j_opt);
  optim_out_opt
;;

let make_final
    final_engine ({ Dbg.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  final_engine spf pm;
;;

let make_tasks task_engines =
  let optim = make_optim task_engines.optim_engine in
  let init = make_init task_engines.init_engine optim in
  let grad = make_grad task_engines.grad_engine in
  let final = make_final task_engines.final_engine in
  { init = init;
    grad = grad;
    optim = optim;
    final = final; }
;;

(* Wrappers. *)

let mk_init_proj { Dbg.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> _1_ then
    failwith "init_proj" "initialization should receive one value" else
  report "one value received";
  let dim = Nat.of_int (Pio_tv.to_int (A_ng.array_get_first vals)) in
  report "dimension extracted";
  dim
;;

let mk_init_inj _spf ((ni, nj), mpm1, (sc1, maa1, j1)) =
  let sz = (Nat.to_int ni, Nat.to_int nj) in
  let sc1 = Os.to_string sc1 in
  let (lens1, a1) = A_ng.array_array_pack maa1 in
  let lens1 = A.map Nat.to_int lens1 in
  [| Pio_tv.of_int_couple sz;
     Pio_tv.of_c_int_matrix mpm1;
     Pio_tv.of_string sc1;
     Pio_tv.of_int_vector lens1;
     Pio_tv.of_float_vector a1;
     Pio_tv.of_float j1; |]
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

let mk_grad_proj { Dbg.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> _1_ then
    failwith "grad_proj" "gradient should receive one value" else
  report "one value received";
  let pm = Pio_tv.to_c_float_matrix (A_ng.array_get_first vals) in
  report "fine parameter extracted";
  pm
;;

let mk_grad_inj _spf gm = [| Pio_tv.of_c_float_matrix gm; |];;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

let mk_optim_proj { Dbg.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> _3_ then
    failwith "optim_proj" "optimization should receive three values" else
  report "three values received";
  let ppm = Pio_tv.to_c_int_matrix (A_ng.array_get_first vals) in
  let ppm = M.map Nat.of_int ppm in
  report "partition extracted";
  let lens_init = Pio_tv.to_int_vector (A.get vals _1_) in
  let lens_init = A.map Nat.of_int lens_init in
  report "coarse parameter lengths extracted";
  let a_init = Pio_tv.to_float_vector (A.get vals _2_) in
  let maa_init = A_ng.array_array_unpack lens_init a_init in
  report "initial coarse parameter extracted";
  (ppm, maa_init)
;;

let mk_optim_inj _spf (sc_opt, maa_opt, j_opt) =
  let sc_opt = Os.to_string sc_opt in
  let (lens_opt, a_opt) = A_ng.array_array_pack maa_opt in
  let lens_opt = A.map Nat.to_int lens_opt in
  [| Pio_tv.of_string sc_opt;
     Pio_tv.of_int_vector lens_opt;
     Pio_tv.of_float_vector a_opt;
     Pio_tv.of_float j_opt; |]
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

let mk_final_proj { Dbg.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> _1_ then
    failwith "final_proj" "finalization should receive one value" else
  report "one value received";
  let pm = Pio_tv.to_c_float_matrix (A_ng.array_get_first vals) in
  report "fine parameter extracted";
  pm
;;

let mk_final_inj _spf _ = [| Pio_tv.of_string "OK, bye!"; |];;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;

(* Task computation functions. *)

let make_initialize init pf piob =
  Pio.worker_wrap_task piob (make_init_wrapper pf "init") init
;;

let make_compute_gradient grad pf piob =
  Pio.worker_wrap_task piob (make_grad_wrapper pf "grad") grad
;;

let make_optimize optim pf piob =
  Pio.worker_wrap_task piob (make_optim_wrapper pf "optim") optim
;;

let make_finalize final pf piob vals =
  Pio.worker_wrap_task piob (make_final_wrapper pf "final") final vals;
  (raise Exit : unit);
;;

(* Main loop. *)

let welcome_action () = "start user model"
and farewell_action () = "finish user model"
;;

let mk_task_tbl tasks =
  L1.of_list
    [
     ("init", make_initialize tasks.init);
     ("grad", make_compute_gradient tasks.grad);
     ("optim", make_optimize tasks.optim);
     ("final", make_finalize tasks.final);
    ]
;;

let main_loop tasks =
  let task_tbl = mk_task_tbl tasks in
  Pio.worker_main welcome_action farewell_action task_tbl
;;
