CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                     C
C                              Ref-indic                              C
C                                                                     C
C                        Refinement indicators                        C
C                                                                     C
C                       EPI Serena, INRIA Paris                       C
C                                                                     C
C  Copyright (c) 2013-2017 INRIA.                                     C
C  Institut National de Recherche en Informatique et en Automatique.  C
C  All rights reserved. This file is distributed only by permission.  C
C                                                                     C
C  Hend Ben Ameur <hbenameur@yahoo.ca>                                C
C  Francois Clement <Francois.Clement@inria.fr>                       C
C  Pierre Weis <Pierre.Weis@inria.fr>                                 C
C                                                                     C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

C     $Id: main.fpp,v 1.11 2017/11/06 09:05:31 fclement Exp $

#include "worker_constants.h"

      program main
C     Main program.

      implicit none

C     External routines.
      external init_engine, grad_engine, optim_engine, final_engine

      call worker_main_body
     +     (init_engine, grad_engine, optim_engine, final_engine)

      end


      subroutine init_engine(mp1, pp1, lens1, m1)
C     Initialization.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Output arguments.
      integer mp1(WORKER_MAX_MSIZE, WORKER_MAX_DIMENSION)
      integer pp1(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)
      integer lens1(WORKER_MAX_DIMENSION)
      FLOAT m1(WORKER_MAX_ZONE_NUMBER)

C     Local variables.
      character sbr*11, msg*MAX_ERROR_MSG_LENGTH
      integer s_len, deb, len, dim_loc, i, j

C     External function.
      logical pio_user_get_debug_mode

      s_len = 11
      sbr = 'init_engine'

      if (.not. ((dim .ge. 1) .or. (dim .le. 2))) then
         call worker_error_unbound_dimension(s_len, sbr, dim)
         return
      endif

      deb = 0
      if (pio_user_get_debug_mode()) deb = 1

      call init_model(deb, dim_loc, size_x, size_y, ier, len, msg)
      if (ier .ne. 0) then
         call worker_error_init_engine(s_len, sbr, len, msg)
         return
      endif

      if (dim_loc .ne. dim) then
         len = 28 + 26 + 1
         write(msg, 1001) dim
         call worker_error_init_engine(s_len, sbr, len, msg)
         return
      endif

 1001 format('worker is not configured for',
     +     ' a parameter of dimension ', i1)

      size = size_x * size_y
      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

C     One zone per channel.
      call worker_trivial_init(WORKER_ONE, pp1, lens1, m1)

      msize = size
      do j = 1, dim
         do i = 1, msize
            mp1(i, j) = pp1(i, j)
         enddo
      enddo

      return
      end


      subroutine grad_engine(p, g)
C     Gradient computation.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Input arguments.
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Output argument.
      FLOAT g(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character sbr*11
      integer s_len, ndS, ndT, i
      real*8 S(np_max), T(np_max), gdS(np_max), gdT(np_max)

      s_len = 11
      sbr = 'grad_engine'

      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

      ndS = size
      ndT = size
      do i = 1, size
         S(i) = p(i, 1)
         T(i) = p(i, 2)
      enddo

      call gradient(ndS, ndT, S, T, gdS, gdT)

      do i = 1, size
         g(i, 1) = gdS(i)
         g(i, 2) = gdT(i)
      enddo

      return
      end


      subroutine optim_engine(pp, lens, m, len_sc, sc, cost)
C     Optimal coarse parameter computation.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Input arguments.
      integer pp(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)
      integer lens(WORKER_MAX_DIMENSION)

C     Input/output argument.
      FLOAT m(WORKER_MAX_ZONE_NUMBER)

C     Output argument.
      character sc*(*)
      integer len_sc
      FLOAT cost

C     Local variables.
      character sbr*12, msg*MAX_ERROR_MSG_LENGTH
      integer s_len, ncells, i, j, len, zone_number_loc
      integer projS(np_max), nzonesS, projT(np_max), nzonesT
      real*8 paramS(np_max), paramT(np_max), cout

      s_len = 12
      sbr = 'optim_engine'

      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

      ncells = size
      do i = 1, ncells
         projS(i) = pp(i, 1)
         projT(i) = pp(i, 2)
      enddo
      nzonesS = lens(1)
      nzonesT = lens(2)
      zone_number_loc = nzonesS + nzonesT

      if (nzonesS .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, nzonesS)
         return
      endif
      if (nzonesT .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, nzonesT)
         return
      endif

      if (zone_number_loc .ne. zone_number) then
         call worker_error_wrong_size(
     +        s_len, sbr, zone_number_loc, zone_number)
         return
      endif

      j = 0
      do i = 1, nzonesS
         j = j + 1
         paramS(i) = m(j)
      enddo
      do i = 1, nzonesT
         j = j + 1
         paramT(i) = m(j)
      enddo

      call pio_user_report_message('before actual optimization')

      call optim(ncells, projS, nzonesS, paramS, projT, nzonesT, paramT,
     +     cout, len_sc, sc, ier, len, msg)
      if (ier .ne. 0) then
         call worker_error_optim_engine(s_len, sbr, len, msg)
         return
      endif

      call pio_user_report_message('after actual optimization')

      j = 0
      do i = 1,nzonesS
         j = j + 1
         m(j) = paramS(i)
      enddo
      do i = 1,nzonesT
         j = j + 1
         m(j) = paramT(i)
      enddo
      cost = FLOAT_OF_DOUBLE(cout)

      return
      end


      subroutine final_engine(p)
C     Finalization.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'

C     Input argument.
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character sbr*12
      integer s_len

      s_len = 12
      sbr = 'final_engine'

      call save_opt_param(s_len, sbr, p)
      if (ier .ne. 0) return

      return
      end


      subroutine save_opt_param(s_len, sbr, p)
C     Save the optimal transmissivity parameter.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'

C     Input argument.
      character sbr*(*)
      integer s_len
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character fname*13
      integer fn_len, i

      fn_len = 13
      fname = 'opt_param.dat'
      open(unit=WORKER_IOUNIT, file=fname(1:fn_len),
     +     status='unknown', access='sequential', form='formatted',
     +     iostat=ier)
      if (ier .ne. 0) then
         call worker_error_open_failure(s_len, sbr, fn_len, fname)
         return
      endif

      write(WORKER_IOUNIT, *) size
      do i = 1, size
         write(WORKER_IOUNIT, *) p(i, 1)
      enddo

      write(WORKER_IOUNIT, *) size
      do i = 1, size
         write(WORKER_IOUNIT, *) p(i, 2)
      enddo

      close(WORKER_IOUNIT)

      return
      end
